/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "video.h"
#include "m6809.h"
#include "sam.h"
#include "vdg.h"
#include "pia.h"
#include "logging.h"
#include "types.h"

#define STATE_LBORDER 0
#define STATE_DISPLAY 1
#define STATE_RBORDER 2
#define STATE_HBLANK 3
#define STATE_VBLANK 4

#define RENDER_ALPHA 0
#define RENDER_G3C 1
#define RENDER_G3R 2
#define RENDER_G6C 3
#define RENDER_G6R 4
Uint32 render_func;
static inline void render_alpha(void);
static inline void render_g3c(void);
static inline void render_g3r(void);
static inline void render_g6c(void);
static inline void render_g6r(void);


Uint32 vdg_gmode;
static Uint32 vdg_css;
static Uint32 vdg_isgraphics;
static Uint32 subline;
static Uint32 longline;
static Uint8 *video_ptrs[2];

extern Uint8 vdg_alpha[768];

PIXEL *pixel;
PIXEL darkgreen, black;
PIXEL bg_colour;
PIXEL fg_colour;
PIXEL vdg_colour[8];
PIXEL *cg_colours;
static PIXEL border_colour;

void vdg_init(void) {
	video_init();
	vdg_colour[0] = MAPCOLOUR(0x00, 0xff, 0x00);
	vdg_colour[1] = MAPCOLOUR(0xff, 0xff, 0x00);
	vdg_colour[2] = MAPCOLOUR(0x00, 0x00, 0xff);
	vdg_colour[3] = MAPCOLOUR(0xff, 0x00, 0x00);
	vdg_colour[4] = MAPCOLOUR(0xff, 0xe0, 0xe0);
	vdg_colour[5] = MAPCOLOUR(0x00, 0xff, 0xff);
	vdg_colour[6] = MAPCOLOUR(0xff, 0x00, 0xff);
	vdg_colour[7] = MAPCOLOUR(0xff, 0xa5, 0x00);
	black = MAPCOLOUR(0x00, 0x00, 0x00);
	darkgreen = MAPCOLOUR(0x00, 0x20, 0x00);
	PIA_1B.wrfunc[0] = vdg_set_mode;
}

void vdg_reset(void) {
	vdg_set_mode(0x00);
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	subline = 0;
}

void vdg_render_scanline(void) {
	Uint32 i;
	for (i = 32; i; i--) {
		*pixel = border_colour;
		*(pixel + 288 * XSTEP) = border_colour;
		pixel += XSTEP;
	}
	/* Fetch current video base addresses */
	sam_video_pointers(video_ptrs);
	switch(render_func) {
		case RENDER_ALPHA: render_alpha(); break;
		case RENDER_G3C: render_g3c(); break;
		case RENDER_G3R: render_g3r(); break;
		case RENDER_G6C: render_g6c(); break;
		case RENDER_G6R: render_g6r(); break;
	}
	pixel += NEXTLINE + (32 * XSTEP);
	subline++;
	if (subline > 11)
		subline = 0;
}

void vdg_vsync(void) {
	VIDEO_UPDATE;
	sam_video_vsync();
	pixel = VIDEO_TOPLEFT + VIDEO_VIEWPORT_YOFFSET;
	subline = 0;
}

static inline void render_alpha(void) {
	Uint8 *vram_ptr;
	Uint32 i, j, octet;
	Uint16 tmp;
	for (j = 0; j <= longline; j++) {
		vram_ptr = video_ptrs[j];
		for (i = 16; i; i--) {
			octet = *(vram_ptr++);
			if (octet & 0x80) {
				tmp = vdg_colour[(octet & 0x70)>>4];
				if (subline < 6)
					octet >>= 2;
				*pixel = *(pixel+1*XSTEP) = *(pixel+2*XSTEP) = *(pixel+3*XSTEP) = (octet & 0x02) ? tmp : black;
				*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = *(pixel+6*XSTEP) = *(pixel+7*XSTEP) = (octet & 0x01) ? tmp : black;
				pixel += 8*XSTEP;
			} else {
				tmp = vdg_alpha[(octet&0x3f)*12 + subline];
				if (octet & 0x40)
					tmp = ~tmp;
				*pixel = (tmp & 0x80) ? fg_colour : bg_colour;
				*(pixel+1*XSTEP) = (tmp & 0x40) ? fg_colour : bg_colour;
				*(pixel+2*XSTEP) = (tmp & 0x20) ? fg_colour : bg_colour;
				*(pixel+3*XSTEP) = (tmp & 0x10) ? fg_colour : bg_colour;
				*(pixel+4*XSTEP) = (tmp & 0x08) ? fg_colour : bg_colour;
				*(pixel+5*XSTEP) = (tmp & 0x04) ? fg_colour : bg_colour;
				*(pixel+6*XSTEP) = (tmp & 0x02) ? fg_colour : bg_colour;
				*(pixel+7*XSTEP) = (tmp & 0x01) ? fg_colour : bg_colour;
				pixel += 8*XSTEP;
			}
		}
	}
}

static inline void render_g3c(void) {
	Uint8 *vram_ptr;
	Uint32 i, j, octet;
	for (j = 0; j <= longline; j++) {
		vram_ptr = video_ptrs[j];
		for (i = 16; i; i--) {
			octet = *(vram_ptr++);
			*pixel = *(pixel+1*XSTEP) = *(pixel+2*XSTEP)
				= *(pixel+3*XSTEP)
				= cg_colours[(octet & 0xc0) >> 6];
			*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = *(pixel+6*XSTEP)
				= *(pixel+7*XSTEP)
				= cg_colours[(octet & 0x30) >> 4];
			*(pixel+8*XSTEP) = *(pixel+9*XSTEP) = *(pixel+10*XSTEP)
				= *(pixel+11*XSTEP)
				= cg_colours[(octet & 0x0c) >> 6];
			*(pixel+12*XSTEP) = *(pixel+13*XSTEP)
				= *(pixel+14*XSTEP) = *(pixel+15*XSTEP)
				= cg_colours[(octet & 0x03) >> 4];
			pixel += 16*XSTEP;
		}
	}
}

static inline void render_g3r(void) {
	Uint8 *vram_ptr;
	Uint32 i, j, octet;
	for (j = 0; j <= longline; j++) {
		vram_ptr = video_ptrs[j];
		for (i = 16; i; i--) {
			octet = *(vram_ptr++);
			*pixel = *(pixel+1*XSTEP) = (octet & 0x80) ? fg_colour : bg_colour;
			*(pixel+2*XSTEP) = *(pixel+3*XSTEP) = (octet & 0x40) ? fg_colour : bg_colour;
			*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = (octet & 0x20) ? fg_colour : bg_colour;
			*(pixel+6*XSTEP) = *(pixel+7*XSTEP) = (octet & 0x10) ? fg_colour : bg_colour;
			*(pixel+8*XSTEP) = *(pixel+9*XSTEP) = (octet & 0x08) ? fg_colour : bg_colour;
			*(pixel+10*XSTEP) = *(pixel+11*XSTEP) = (octet & 0x04) ? fg_colour : bg_colour;
			*(pixel+12*XSTEP) = *(pixel+13*XSTEP) = (octet & 0x02) ? fg_colour : bg_colour;
			*(pixel+14*XSTEP) = *(pixel+15*XSTEP) = (octet & 0x01) ? fg_colour : bg_colour;
			pixel += 16*XSTEP;
		}
	}
}

static inline void render_g6c(void) {
	Uint8 *vram_ptr;
	Uint32 i, j, octet;
	for (j = 0; j <= longline; j++) {
		vram_ptr = video_ptrs[j];
		for (i = 16; i; i--) {
			octet = *(vram_ptr++);
			*pixel = *(pixel+1*XSTEP) = cg_colours[(octet & 0xc0) >> 6];
			*(pixel+2*XSTEP) = *(pixel+3*XSTEP) = cg_colours[(octet & 0x30) >> 4];
			*(pixel+4*XSTEP) = *(pixel+5*XSTEP) = cg_colours[(octet & 0x0c) >> 2];
			*(pixel+6*XSTEP) = *(pixel+7*XSTEP) = cg_colours[octet & 0x03];
			pixel += 8*XSTEP;
		}
	}
}

static inline void render_g6r(void) {
	Uint8 *vram_ptr;
	Uint32 i, j, octet;
	for (j = 0; j <= longline; j++) {
		vram_ptr = video_ptrs[j];
		for (i = 16; i; i--) {
			octet = *(vram_ptr++);
			*pixel = (octet & 0x80) ? fg_colour : bg_colour;
			*(pixel+1*XSTEP) = (octet & 0x40) ? fg_colour : bg_colour;
			*(pixel+2*XSTEP) = (octet & 0x20) ? fg_colour : bg_colour;
			*(pixel+3*XSTEP) = (octet & 0x10) ? fg_colour : bg_colour;
			*(pixel+4*XSTEP) = (octet & 0x08) ? fg_colour : bg_colour;
			*(pixel+5*XSTEP) = (octet & 0x04) ? fg_colour : bg_colour;
			*(pixel+6*XSTEP) = (octet & 0x02) ? fg_colour : bg_colour;
			*(pixel+7*XSTEP) = (octet & 0x01) ? fg_colour : bg_colour;
			pixel += 8*XSTEP;
		}
	}
}

void vdg_set_mode(Uint32 mode) {
	vdg_css = (mode & 0x08) >> 1;
	vdg_gmode = (mode & 0x70) >> 4;
	vdg_isgraphics = mode & 0x80;
	if (vdg_isgraphics) {
		cg_colours = &vdg_colour[vdg_css];
		bg_colour = black;
		fg_colour = vdg_colour[vdg_css];
		border_colour = fg_colour;
		switch (vdg_gmode) {
			case 0: longline = 0; render_func = RENDER_G3C; break;
			case 1:
			case 3:
			case 5: longline = 0; render_func = RENDER_G3R; break;
			case 2:
			case 4:
			case 6: longline = 1; render_func = RENDER_G6C; break;
			default:
			case 7: longline = 1; render_func = RENDER_G6R; break;
		}
	} else {
		longline = 1;
		bg_colour = darkgreen;
		border_colour = black;
		if (vdg_css)
			fg_colour = vdg_colour[7];
		else
			fg_colour = vdg_colour[0];
		render_func = RENDER_ALPHA;
	}
}
